// Copyright (c) Lawrence Livermore National Security, LLC and other Conduit
// Project developers. See top-level LICENSE AND COPYRIGHT files for dates and
// other details. No copyright assignment is required to contribute to Conduit.

//-----------------------------------------------------------------------------
///
/// file: conduit_config.h
///
//-----------------------------------------------------------------------------

#ifndef CONDUIT_CONFIG_H
#define CONDUIT_CONFIG_H

//-----------------------------------------------------------------------------
//
// #define platform check helpers
//
//-----------------------------------------------------------------------------

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32) && !defined(__CYGWIN__)
#define CONDUIT_PLATFORM_WINDOWS
#elif  defined(__APPLE__)
#define CONDUIT_PLATFORM_APPLE
#else
#define CONDUIT_PLATFORM_UNIX
#endif

/* #undef CONDUIT_INSTALL_PREFIX */

// version info
#define CONDUIT_VERSION "0.9.5"
#define CONDUIT_VERSION_MAJOR 0
#define CONDUIT_VERSION_MINOR 9
#define CONDUIT_VERSION_PATCH 5

#define CONDUIT_GIT_SHA1 "8543892a582737eb5941b1e90df068a75543df92"
/* #undef CONDUIT_GIT_SHA1_ABBREV */
/* #undef CONDUIT_GIT_TAG */

#define CONDUIT_SYSTEM_TYPE "Linux-6.18.11-1-armv7-ARCH"

#define CONDUIT_CPP_COMPILER "/usr/lib/distcc/bin/c++"

/* #undef CONDUIT_FORTRAN_COMPILER */

/* #undef CONDUIT_USE_FMT */

// keep CONDUIT_USE_CXX11 to support old logic
#define CONDUIT_USE_CXX11

#define CONDUIT_USE_CXX14

#define CONDUIT_USE_YYJSON

/* #undef CONDUIT_USE_CALIPER */

/* #undef CONDUIT_USE_OPENMP */

/* #undef CONDUIT_USE_PARMETIS */

/* #undef CONDUIT_USE_TOTALVIEW */


#endif


