/*
 * Distributed under OSI-approved BSD 3-Clause License. See
 * accompanying License.txt
 */

#ifndef catalyst_hpp
#define catalyst_hpp

#include "catalyst.h"
#include "catalyst_conduit.hpp"
#include "catalyst_conduit_error.hpp"

#endif // #ifndef catalyst_hpp
